# Copyright (c) 2013, 2014, 2015 MIT Probabilistic Computing Project.
#
# This file is part of Venture.
#
# Venture is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Venture is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Venture.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from venture.lite.exception import VentureError
from venture.lite.value import VentureValue
from venture.lite.value import registerVentureType
# Used by the exec pylint: disable=unused-import
from venture.lite.types import VentureType
from venture.lite.types import standard_venture_type
import venture.lite.typing as tp

T = tp.TypeVar('T')

# Environments store Python strings for the symbols, not Venture
# symbol objects.  This is a choice, but whichever way it is made it
# should be consistent.
# Binding a symbol to the Python value None has the effect of making
# the symbol behave as if unbound, shadowing existing bindings; this
# behavior is used in the implementation of letrec.
class VentureEnvironment(VentureValue, tp.Generic[T]):
  __slots__ = ('outerEnv', 'frame')
  def __init__(self,outerEnv=None,ids=None,nodes=None):
    # type: (VentureEnvironment, tp.List[str], tp.List[T]) -> None
    self.outerEnv = outerEnv
    self.frame = OrderedDict() # type: OrderedDict[str, T]
    if ids is not None:
      for sym in ids:
        assert isinstance(sym, str)
    if ids is not None and nodes is not None:
      self.frame.update(zip(ids,nodes))

  def addBinding(self,sym,val):
    # type: (str, T) -> None
    if not isinstance(sym, str):
      raise VentureError("Symbol '%s' must be a string, not %s" % (str(sym), type(sym)))
    if sym in self.frame:
      raise VentureError("Symbol '%s' already bound" % sym)
    self.frame[sym] = val

  def removeBinding(self,sym):
    assert isinstance(sym, str)
    if sym in self.frame: del self.frame[sym]
    elif not self.outerEnv: raise VentureError("Cannot unbind unbound symbol '%s'" % sym)
    else: self.outerEnv.removeBinding(sym)

  def fillBinding(self,sym,val):
    # Used in the implementation of letrec
    assert isinstance(sym, str)
    assert sym in self.frame
    assert self.frame[sym] is None
    self.frame[sym] = val

  def findSymbol(self,sym):
    # type: (str) -> T
    if sym in self.frame: ret = self.frame[sym]
    elif not self.outerEnv: ret = None
    else: ret = self.outerEnv.findSymbol(sym)
    if ret is None:
      raise VentureError("Cannot find symbol '%s'" % sym)
    return ret

  def symbolBound(self, sym):
    if sym in self.frame: return self.frame[sym] is not None
    elif not self.outerEnv: return False
    else: return self.outerEnv.symbolBound(sym)

  # VentureEnvironments are intentionally not comparable until we
  # decide otherwise

  def getEnvironment(self): return self

  def asStackDict(self, _trace=None):
    # Methinks environments can be pretty opaque things for now.
    return OrderedDict([("type", "environment"), ("value", self)])
  @staticmethod
  def fromStackDict(thing): return thing["value"]

  def equalSameType(self, other):
    # This compares node identities, not their contents.  This is as
    # it should be, because nodes can mutate.
    if self.frame == other.frame:
      if self.outerEnv is None:
        return other.outerEnv is None
      elif other.outerEnv is None:
        return False
      else:
        return self.outerEnv.equalSameType(other.outerEnv)
    else: return False

  def lookup(self, key):
    return self.findSymbol(key.getSymbol())
  # TODO Define contains to check whether the symbol is there (without throwing exceptions)

  def printEnv(self):
    print "-- Frame --"
    for (k, v) in self.frame.iteritems():
      print k, v
    if self.outerEnv is not None and self.outerEnv.outerEnv is not None:
      # Skip the top frame, which is presumably the global environment
      self.outerEnv.printEnv()

registerVentureType(VentureEnvironment, "environment")
# Exec is appropriate for metaprogramming
exec(standard_venture_type("Environment", value_classname="VentureEnvironment")) # pylint: disable=exec-used
